/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package bia.moea;

import bia.tsp.Node;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import javax.swing.JFrame;

/**
 * @author Jiri Kubalik
 */
public class NSGA2Knapsack {

    //--- Input data.
    static File dataset;

    /**
     * Old and new population.
     */
    static ArrayList<KnapsackIndividual> oldPop;
    static ArrayList<KnapsackIndividual> newPop;

    /**
     * Newly generated individuals
     */
    static KnapsackIndividual[] offspringPop;

    /**
     * NSGA parameters
     */
    public static KnapsackParameters parameters = KnapsackParameters.getInstance();

    /**
     * Main function.
     *
     * @param args the command line arguments
     */
    public static void main(String[] args) throws Exception {

        parameters.readConfigFile();
        
        dataset = new File(parameters.getDataset() + ".data");
        if (dataset.exists()) {
            importDatasetKnapsack();
        }
        else{
            System.out.println("Dataset file does not exist!");
            System.exit(0);
        }
        
        run();
    }

    public static void run() throws Exception {

        int totEvals = 0;

        //--- Allocate memory for arrays
        oldPop = new ArrayList<KnapsackIndividual>();
        offspringPop = new KnapsackIndividual[2];

        //--- Initialise population.
        totEvals = initializePopulation(oldPop);
        System.out.println("\n\nInitial non-dominated front");
        System.out.println("===========================");
        ArrayList<KnapsackIndividual> initialFront = printNonDominatedSolutions(oldPop);

        //--- Evolve population using either generational or steady-state replacement strategy
        oldPop = evolvePopulation(oldPop, totEvals);

        //--- Find and print non-dominated solutions
        System.out.println("\n\nFinal non-dominated front");
        System.out.println("=========================");
        ArrayList<KnapsackIndividual> finalFront = printNonDominatedSolutions(oldPop);

        //--- Visualize non-dominated fronts
        KnapsackViz.visualizeNonDominatedFronts(initialFront, finalFront);
    }

    /**
     * Evolves the @param population either using a generational or steady-state replacement strategy.
     *
     * @param population the population to be evolved.
     * @param totEvals   a number of fitness function evaluations that have been calculated
     *                   during an initialization phase.
     */
    public static ArrayList<KnapsackIndividual> evolvePopulation(ArrayList<KnapsackIndividual> population, int totEvals) {
        int nbGenerations = 0;
        newPop = new ArrayList<KnapsackIndividual>();

        //--- Find fronts in the current population
        ArrayList<ArrayList<KnapsackIndividual>> fronts = findFronts(population);

        //--- Calculate crowding distance of each individual
        calculateCrowdingDistances(fronts);

        //--- Create population
        population = createNewPopulation(fronts);
        
        //--- While the termination criterion is not satisfied ...
        while (totEvals < parameters.getEvaluations()) {
            //--- Create intermediate population
            newPop.clear();
            while(newPop.size() < parameters.getPopSize()) {
                int[] par = selection(population);
                KnapsackIndividual parent1 = population.get(par[0]).clone();
                KnapsackIndividual parent2 = population.get(par[1]).clone();
                //--- Create two new individuals.
                if (Math.random() < parameters.getPC()) {
                    offspringPop = parent1.crossover(parent2);
                    offspringPop[0].mutation();
                    offspringPop[1].mutation();
                } else {
                    offspringPop[0] = parent1;
                    offspringPop[1] = parent2;
                    offspringPop[0].mutation();
                    offspringPop[1].mutation();
                }
                //--- Calculates objective values of new individuals.
                offspringPop[0].getObjectives();
                offspringPop[1].getObjectives();
                totEvals += 2;  // Increase number of evaluations done.

                //--- Put offspring to intermediate population.
                newPop.add(offspringPop[0]);
                if(newPop.size() < parameters.getPopSize())
                    newPop.add(offspringPop[1]);
            }

            //--- Merge population and newPop.
            for(KnapsackIndividual ind : newPop)
                population.add(ind);

            //--- Update the population
            //--- Find fronts in the current population
            fronts = findFronts(population);
            
            //--- Calculate crowding distance of each individual
            calculateCrowdingDistances(fronts);

            //--- Create population
            population = createNewPopulation(fronts);
            
            System.out.print(".");
        }
        //---
        return population;        
    }

    /**
     * Initialization of population.
     *
     * @return a number of fitness evaluations calculated during the population initialization
     */
    public static int initializePopulation(ArrayList<KnapsackIndividual> population) {

        population.clear();
        // Generate individuals in parent population and calculate their fitness values.
        for (int ind = 0; ind < parameters.getPopSize(); ind++) {
            KnapsackIndividual newInd = new KnapsackIndividual();
            newInd.getObjectives(); // Calculates objective values
            population.add(newInd);
        }

        return population.size();
    }

    /**
     * Splits the population into non-dominated fronts.
     * Each individual is assigned a value 'frontId' that takes values 0...N,
     * where lower value represents better front.
     *
     */
    public static ArrayList<ArrayList<KnapsackIndividual>> findFronts(ArrayList<KnapsackIndividual> population){
        ArrayList<ArrayList<KnapsackIndividual>> fronts = new ArrayList();
        int frontNb = 0;

        //--- Divide the whole population into non-dominated fronts
        while(!population.isEmpty()){
            ArrayList<KnapsackIndividual> newFront = new ArrayList<KnapsackIndividual>();
            ArrayList newFrontIndexes = new ArrayList();
            int origSize = population.size();   //--- check all remaining individuals
            for(int i = 0; i < origSize; i++){
                boolean dominated = false;
                KnapsackIndividual candidate = population.get(i);
                for(KnapsackIndividual ind : population){
                    if(ind.dominates(candidate)){
                        dominated = true;
                        break;
                    }
                }
                if(!dominated)
                    newFrontIndexes.add(i);
            }
            //--- Fill the newFront with the non-dominated individuals
            for(int i = newFrontIndexes.size()-1; i >= 0; i--){
                KnapsackIndividual ind = population.remove((int)newFrontIndexes.get(i));
                ind.frontId = frontNb;
                newFront.add(ind);    //--- put non-dominated candidate to the newFront
            }
            //--- Add newFront into fronts
            fronts.add(newFront);
            frontNb++;
        }

        return fronts;
    }

    /**
     * Calculates a crowding distance of each individual in the population.
     *
     */
    public static void calculateCrowdingDistances(ArrayList<ArrayList<KnapsackIndividual>> fronts){
        
        //--- Process population front by front
        for(ArrayList<KnapsackIndividual> front : fronts){
            for(int n = 0; n < 2; n++){
                //--- Sort the individuals w.r.t. n-th objective, objectives[n] ...
                if(n == 0)
                    Collections.sort(front, new Comparator<KnapsackIndividual>() {
                        public int compare(KnapsackIndividual a, KnapsackIndividual b) {
                            if(a.objectives[0] == b.objectives[0])
                                return 0;
                            else
                                return (a.objectives[0] > b.objectives[0]) ? 1 : -1;
                        }
                    });
                else
                    Collections.sort(front, new Comparator<KnapsackIndividual>() {
                        public int compare(KnapsackIndividual a, KnapsackIndividual b) {
                            if(a.objectives[1] == b.objectives[1])
                                return 0;
                            else
                                return (a.objectives[1] > b.objectives[1]) ? 1 : -1;
                        }
                    });
                //--- ... and calculate distances between nearest neighbors
                for(int i = 0; i < front.size(); i++){
                    if((i == 0) || (i == front.size()-1))
                        front.get(i).setCrowdingDist(100000); //--- individuals with extreme objective values have maximal crowding distance
                    else{
                        front.get(i).setCrowdingDist(front.get(i).crowdingDist + Math.abs(front.get(i+1).objectives[n] - front.get(i-1).objectives[n]));
                    }
                }
            }
        }
    }
    
    /**
     * Creates a new population using an elitist replacement strategy.
     * Better fronts are preferred; within a particular front, solutions with bigger crowding distance are preferred.
     */
    public static ArrayList<KnapsackIndividual> createNewPopulation(ArrayList<ArrayList<KnapsackIndividual>> fronts){
        ArrayList<KnapsackIndividual> population = new ArrayList<KnapsackIndividual>();

        while(population.size() < parameters.getPopSize()){
            ArrayList<KnapsackIndividual> currFront = fronts.remove(0); //--- best front first
            if(parameters.getPopSize() - population.size() >= currFront.size()){
                //--- Put all individuals of the current front to the population
                for(KnapsackIndividual ind : currFront)
                    population.add(ind);
            }
            else{
                //--- Sort individuals of the current front w.r.t. crowding distance ...
                Collections.sort(currFront, new Comparator<KnapsackIndividual>() {
                    public int compare(KnapsackIndividual a, KnapsackIndividual b) {
                        if(a.crowdingDist == b.crowdingDist)
                            return 0;
                        else
                            return (a.crowdingDist < b.crowdingDist) ? 1 : -1;
                    }
                });
                //--- ... and take the first (parameters.getPopSize() - population.size()) individuals
                int howMany = parameters.getPopSize() - population.size();
                for(int i = 0; i < howMany; i++)
                    population.add(currFront.get(i));
            }
        }
        
        return population;
    }

    /**
     * Prints the set of non-dominated solutions of the population, 
     * i.e. individuals with 'frontId' equal to 0.
     *
     */
    public static ArrayList<KnapsackIndividual> printNonDominatedSolutions(ArrayList<KnapsackIndividual> population){
        ArrayList<KnapsackIndividual> front = new ArrayList<KnapsackIndividual>();

        //--- Select non-dominated individuals from the population.
        for(int i = 0; i < population.size(); i++){
            boolean dominated = false;
            KnapsackIndividual candidate = population.get(i);
            for(KnapsackIndividual ind : population){
                if(ind.dominates(candidate)){
                    dominated = true;
                    break;
                }
            }
            if(!dominated)
                front.add(candidate);    //--- put non-dominated candidate to the front
        }
        
        //--- Print them out.
        System.out.println("Nondominated front:");
        for(KnapsackIndividual ind : front){
            System.out.println(ind.objectives[0] + "  " + ind.objectives[1]);
        }
        
        return front;
    }

    /**
     * Tournament selection - select a winner of one tournament between "parameters.getTournamentSize()"
     *    randomly chosen individuals from the population of size "parameters.getPopSize()".
     * It performs binary comparisons of two individuals, A and B, that determine a winner 
     * according to the following rule:
     *    A is better than B iff
     *       1) A is from a better front than B
     *       or
     *       2) A is from the same from as B and has bigger crowding distance value.
     *
     * @return index of winner individual.
     */
    public static int tournamentSelection(ArrayList<KnapsackIndividual> p) {
        /* PUT YOUR CODE HERE */
        
        int winnerIndex = (int) (Math.random() * (parameters.getPopSize())); // First member of tournament selection.
        int winnerFront = p.get(winnerIndex).frontId;
        int winnerCrowdingDist = p.get(winnerIndex).crowdingDist;

        // Choose N individuals and select best one.
        // We already chose randomly first individual 3 rows up, so there is only SGA.tournament_size-1 random individuals to check. 
        for (int individualIndex = 0; individualIndex < parameters.getTournamentSize() - 1; individualIndex++) {
            int index = (int) (Math.random() * (parameters.getPopSize()));
            if ((p.get(index).frontId < winnerFront) ||     //--- candidate is from better front
                ((p.get(index).frontId == winnerFront) && (p.get(index).crowdingDist > winnerCrowdingDist))){   //--- candidate has better crowding distance
                winnerFront = p.get(index).frontId;
                winnerCrowdingDist = p.get(index).crowdingDist;
                winnerIndex = index;
            }
        }

        return winnerIndex;
    }
    
    /**
     * Performs selection method according selectionType value.
     *
     * @return array of two indexes selected by selection methods.
     */
    public static int[] selection(ArrayList<KnapsackIndividual> p) {
        int[] chosen = new int[2];

        chosen[0] = tournamentSelection(p);
        chosen[1] = chosen[0];
        while (chosen[0] == chosen[1])
            chosen[1] = tournamentSelection(p);

        return chosen;
    }    
    
    /**
     * Imports data from specified input file.
     * 
     */
    private static void importDatasetKnapsack() {
        BufferedReader br;
        String temp;

        try {
            br = new BufferedReader(new FileReader(dataset));
            temp = br.readLine();   //--- reads first line "knapsack problem specification (...)"
            String[] line = temp.split(",");
            temp = line[1];
            line = temp.split("\\s++");
            parameters.setNbItems(Integer.parseInt(line[1]));

            KnapsackIndividual.items = new KnapsackItem[parameters.getNbItems()];
            for(int j = 0; j < parameters.getNbItems(); j++){
                KnapsackIndividual.items[j] = new KnapsackItem();
            }
            for(int i = 0; i < 2; i++){
                temp = br.readLine();   //--- reads "="
                temp = br.readLine();   //--- reads second line "knapsack 1:"
                temp = br.readLine();   //--- reads third line "capacity: ..."
                line = temp.split("\\+");
                KnapsackIndividual.capacities[i] = Integer.parseInt(line[1]);
                for(int j = 0; j < parameters.getNbItems(); j++){
                    temp = br.readLine();   //--- reads "item1:"
                    temp = br.readLine();   //--- reads "weight: ..."
                    line = temp.split("\\+");
                    KnapsackIndividual.items[j].setWeight(i, Integer.parseInt(line[1]));
                    //---
                    temp = br.readLine();   //--- reads "profit: ..."
                    line = temp.split("\\+");
                    KnapsackIndividual.items[j].setProfit(i, Integer.parseInt(line[1]));
                    //--- set wpRatio
                    KnapsackIndividual.items[j].calculateWPRatio(i);
                }
            }
            br.close();

            //--- Calculate WeightProfitOrder
            KnapsackIndividual.WeightProfitOrder = new int[2][];     //--- one order for each knapsack
            KnapsackIndividual.WeightProfitOrder[0] = new int[parameters.getNbItems()];
            KnapsackIndividual.WeightProfitOrder[1] = new int[parameters.getNbItems()];
            for(int i = 0; i < parameters.getNbItems(); i++){
                KnapsackIndividual.WeightProfitOrder[0][i] = i;
                KnapsackIndividual.WeightProfitOrder[1][i] = i;
            }
            //--- Sort items in descending order, from the least valuable to the most valuable one
            for(int i = 0; i < parameters.getNbItems(); i++){
                //--- 1st knapsack
                for(int j = 0; j < parameters.getNbItems()-1-i; j++){
                    if(KnapsackIndividual.items[KnapsackIndividual.WeightProfitOrder[0][j]].getWPRatio(0) < KnapsackIndividual.items[KnapsackIndividual.WeightProfitOrder[0][j+1]].getWPRatio(0)){
                        int k = KnapsackIndividual.WeightProfitOrder[0][j];
                        KnapsackIndividual.WeightProfitOrder[0][j] = KnapsackIndividual.WeightProfitOrder[0][j+1];
                        KnapsackIndividual.WeightProfitOrder[0][j+1] = k;
                    }
                }
                //--- 2nd knapsack
                for(int j = 0; j < parameters.getNbItems()-1-i; j++){
                    if(KnapsackIndividual.items[KnapsackIndividual.WeightProfitOrder[1][j]].getWPRatio(1) < KnapsackIndividual.items[KnapsackIndividual.WeightProfitOrder[1][j+1]].getWPRatio(1)){
                        int k = KnapsackIndividual.WeightProfitOrder[1][j];
                        KnapsackIndividual.WeightProfitOrder[1][j] = KnapsackIndividual.WeightProfitOrder[1][j+1];
                        KnapsackIndividual.WeightProfitOrder[1][j+1] = k;
                    }
                }
            }
        } catch (IOException e) {
            System.out.println(e.getMessage());
        }
    }
}
